@@ -1,7 +1,7 @@
 /* Trampoline construction */
 
 /*
- * Copyright 1995-1999, 2001-2003 Bruno Haible, <haible@clisp.cons.org>
+ * Copyright 1995-1999, 2001-2006 Bruno Haible, <bruno@clisp.org>
  *
  * This is free software distributed under the GNU General Public Licence
  * described in the file COPYING. Contact the author if you don't have this
@@ -20,13 +20,17 @@
 #define __hppanew__  /* New trampoline, just a closure. */
 #endif
 #endif
-#if defined(__rs6000__)
+#if defined(__powerpc__) && !defined(__powerpc64__)
 #if !defined(_AIX)
-#define __rs6000sysv4__  /* SysV.4 ABI, real machine code. */
+#define __powerpcsysv4__  /* SysV.4 ABI, real machine code. */
 #else
-#define __rs6000aix__  /* AIX ABI, just a closure. */
+#define __powerpcaix__  /* AIX ABI, just a closure. */
 #endif
 #endif
+#if defined(__powerpc64__)
+/* The only ABI on powerpc64 known so far is the AIX ABI. */
+#define __powerpc64aix__  /* AIX ABI, just a closure. */
+#endif
 #if defined(__hppanew__)
 /*
  * A function pointer is a biased pointer to a data area whose first word
@@ -40,7 +44,7 @@ extern void tramp (); /* trampoline prototype */
 #define CODE_EXECUTABLE
 #endif
 #endif
-#if defined(__rs6000aix__) || defined(__ia64__)
+#if defined(__powerpcaix__) || defined(__powerpc64aix__) || defined(__ia64__)
 /*
  * A function pointer is a pointer to a data area whose first word contains
  * the actual address of the function.
@@ -100,10 +104,8 @@ extern void (*tramp) (); /* trampoline prototype */
 #ifdef HAVE_GETPAGESIZE
 #ifdef __cplusplus
 extern "C" RETGETPAGESIZETYPE getpagesize (void);
-#elif defined(__STDC__)
-extern RETGETPAGESIZETYPE getpagesize (void);
 #else
-extern RETGETPAGESIZETYPE getpagesize ();
+extern RETGETPAGESIZETYPE getpagesize (void);
 #endif
 #else
 #ifdef HAVE_SYS_PARAM_H
@@ -202,7 +204,7 @@ extern RETGETPAGESIZETYPE getpagesize ();
 #include <sys/syslocal.h>
 #endif
 /* Inline assembly function for instruction cache flush. */
-#if defined(__sparc__) || defined(__sparc64__) || defined(__alpha__) || defined(__hppaold__) || defined(__rs6000sysv4__) || defined(__convex__)
+#if defined(__sparc__) || defined(__sparc64__) || defined(__alpha__) || defined(__hppaold__) || defined(__powerpcsysv4__) || defined(__convex__)
 #ifdef __GNUC__
 extern inline
 #if defined(__sparc__) || defined(__sparc64__)
@@ -214,8 +216,8 @@ extern inline
 #ifdef __hppa__
 #include "cache-hppa.c"
 #endif
-#ifdef __rs6000__
-#include "cache-rs6000.c"
+#if defined(__powerpc__) && !defined(__powerpc64__)
+#include "cache-powerpc.c"
 #endif
 #ifdef __convex__
 #include "cache-convex.c"
@@ -279,14 +281,18 @@ extern void __TR_clear_cache();
 #define TRAMP_LENGTH 44
 #define TRAMP_ALIGN 4
 #endif
-#ifdef __rs6000sysv4__
+#ifdef __powerpcsysv4__
 #define TRAMP_LENGTH 36
 #define TRAMP_ALIGN 4
 #endif
-#ifdef __rs6000aix__
+#ifdef __powerpcaix__
 #define TRAMP_LENGTH 24
 #define TRAMP_ALIGN 4
 #endif
+#ifdef __powerpc64aix__
+#define TRAMP_LENGTH 48
+#define TRAMP_ALIGN 8
+#endif
 #ifdef __m88k__
 #define TRAMP_LENGTH 32
 #define TRAMP_ALIGN 8
@@ -299,6 +305,10 @@ extern void __TR_clear_cache();
 #define TRAMP_LENGTH 40
 #define TRAMP_ALIGN 16
 #endif
+#ifdef __x86_64__
+#define TRAMP_LENGTH 32
+#define TRAMP_ALIGN 16
+#endif
 #ifdef __s390__
 #define TRAMP_LENGTH 36
 #define TRAMP_ALIGN 2
@@ -316,14 +326,7 @@ extern void __TR_clear_cache();
 static char* freelist = NULL;
 #endif
 
-#if defined(__STDC__) || defined(__GNUC__) || defined(__cplusplus)
 __TR_function alloc_trampoline (__TR_function address, void* variable, void* data)
-#else
-__TR_function alloc_trampoline (address, variable, data)
-  __TR_function address;
-  void* variable;
-  void* data;
-#endif
 {
   char* function;
 
@@ -511,53 +514,6 @@ __TR_function alloc_trampoline (address, variable, data)
 #define tramp_data(function)  \
   *(unsigned int *) (function +28)
 #endif
-#ifdef __s390__
-  /* function:
-
-        lr      %r0,%r13
-        bras    %r13,.LTN0_0
-.LT0_0:
-.LC0:
-        .long   0x73554711
-.LC1:
-        .long   0x12345678
-.LC2:
-        .long   0xbabebec0
-.LTN0_0:
-        l       %r1,.LC0-.LT0_0(%r13)
-        mvc     0(4,%r1),.LC1-.LT0_0(%r13)
-        l       %r1,.LC2-.LT1_0(%r13)
-        lr      %r13,%r0
-        br      %r1
-  */
-  /* What about big endian / little endian ?? */
-  *(unsigned short *) (function + 0) = 0x180D;
-  *(unsigned int *)   (function + 2) = 0xA7D50008;
-  *(unsigned int *)   (function + 6) = (unsigned int) variable;
-  *(unsigned int *)   (function +10) = (unsigned int) data;
-  *(unsigned int *)   (function +14) = (unsigned int) address;
-  *(unsigned int *)   (function +18) = 0x5810D000;
-  *(unsigned int *)   (function +22) = 0xD2031000;
-  *(unsigned short *) (function +26) = 0xD004;
-  *(unsigned int *)   (function +28) = 0x5810D008;
-  *(unsigned short *) (function +32) = 0x18D0;
-  *(unsigned short *) (function +34) = 0x07f1;
-#define is_tramp(function)  \
-  *(short *)          (function + 0) == 0x180D && \
-  *(int *)            (function + 2) == 0xA7D50008 && \
-  *(int *)            (function +18) == 0x5810D000 && \
-  *(int *)            (function +22) == 0xD2031000 && \
-  *(short *)          (function +26) == 0xD004 && \
-  *(int *)            (function +28) == 0x5810D008 && \
-  *(short *)          (function +32) == 0x18D0 && \
-  *(short *)          (function +34) == 0x07f1
-#define tramp_address(function)  \
-  *(unsigned int *) (function +14)
-#define tramp_variable(function)  \
-  *(unsigned int *) (function +6)
-#define tramp_data(function)  \
-  *(unsigned int *) (function +10)
-#endif
 #ifdef __mips64old__
   /* function:
    *    dli $2,<variable>		3C 02 hi16(hi32(<variable>))
@@ -940,7 +896,7 @@ __TR_function alloc_trampoline (address, variable, data)
 #define tramp_data(function)  \
   ((long *) function)[8]
 #endif
-#ifdef __rs6000sysv4__
+#ifdef __powerpcsysv4__
   /* function:
    *    {liu|lis} 11,hi16(<variable>)		3D 60 hi16(<variable>)
    *    {oril|ori} 11,11,lo16(<variable>)	61 6B lo16(<variable>)
@@ -986,7 +942,7 @@ __TR_function alloc_trampoline (address, variable, data)
 #define tramp_data(function)  \
   hilo(*(unsigned short *) (function +10), *(unsigned short *) (function +14))
 #endif
-#ifdef __rs6000aix__
+#ifdef __powerpcaix__
   /* function:
    *    .long .tramp
    *    .long .mytoc
@@ -1011,6 +967,31 @@ __TR_function alloc_trampoline (address, variable, data)
 #define tramp_data(function)  \
   ((long *) function)[4]
 #endif
+#ifdef __powerpc64aix__
+  /* function:
+   *    .quad .tramp
+   *    .quad .mytoc
+   *    .quad 0
+   * .mytoc:
+   *    .quad <variable>
+   *    .quad <data>
+   *    .quad <address>
+   */
+  *(long *)  (function + 0) = ((long *) &tramp)[0];
+  *(long *)  (function + 8) = (long) (function + 24);
+  *(long *)  (function +16) = 0;
+  *(long *)  (function +24) = (long) variable;
+  *(long *)  (function +32) = (long) data;
+  *(long *)  (function +40) = (long) address;
+#define is_tramp(function)  \
+  ((long *) function)[0] == ((long *) &tramp)[0]
+#define tramp_address(function)  \
+  ((long *) function)[5]
+#define tramp_variable(function)  \
+  ((long *) function)[3]
+#define tramp_data(function)  \
+  ((long *) function)[4]
+#endif
 #ifdef __m88k__
   /* function:
    *    or.u    #r13,#r0,hi16(<variable>)	5D A0 hi16(<variable>)
@@ -1103,6 +1084,93 @@ __TR_function alloc_trampoline (address, variable, data)
 #define tramp_data(function)  \
   ((long *) function)[4]
 #endif
+#ifdef __x86_64__
+  /* function:
+   *    movabsq $<data>,%rax		48 B8 <data>
+   *    movabsq %rax, <variable>	48 A3 <variable>
+   *    movabsq $<address>,%rax		48 B8 <address>
+   *    jmp *%rax			FF E0
+   */
+  *(short *) (function + 0) = 0xB848;
+  *(short *) (function + 2) = (unsigned long) data & 0xffff;
+  *(int *)   (function + 4) = ((unsigned long) data >> 16) & 0xffffffff;
+  *(short *) (function + 8) = ((unsigned long) data >> 48) & 0xffff;
+  *(short *) (function +10) = 0xA348;
+  *(int *)   (function +12) = (unsigned long) variable & 0xffffffff;
+  *(int *)   (function +16) = ((unsigned long) variable >> 32) & 0xffffffff;
+  *(short *) (function +20) = 0xB848;
+  *(short *) (function +22) = (unsigned long) address & 0xffff;
+  *(int *)   (function +24) = ((unsigned long) address >> 16) & 0xffffffff;
+  *(short *) (function +28) = ((unsigned long) address >> 48) & 0xffff;
+  *(short *) (function +30) = 0xE0FF;
+#define is_tramp(function)  \
+  *(unsigned short *) (function + 0) == 0xB848 && \
+  *(unsigned short *) (function +10) == 0xA348 && \
+  *(unsigned short *) (function +20) == 0xB848 && \
+  *(unsigned short *) (function +30) == 0xE0FF
+#define hilo(hiword,loword)  \
+  (((unsigned long) (hiword) << 32) | (unsigned long) (loword))
+#define himidlo(hishort,midword,loshort)  \
+  (((unsigned long) (hishort) << 48) | (unsigned long) (midword) << 16 \
+   | (unsigned long) (loshort))
+#define tramp_address(function)  \
+  himidlo(*(unsigned short *) (function +28), \
+          *(unsigned int *)   (function +24), \
+          *(unsigned short *) (function +22))
+#define tramp_variable(function)  \
+  hilo(*(unsigned int *) (function +16), *(unsigned int *) (function +12))
+#define tramp_data(function)  \
+  himidlo(*(unsigned short *) (function + 8), \
+          *(unsigned int *)   (function + 4), \
+          *(unsigned short *) (function + 2))
+#endif
+#ifdef __s390__
+  /* function:
+
+        lr      %r0,%r13
+        bras    %r13,.LTN0_0
+.LT0_0:
+.LC0:
+        .long   0x73554711
+.LC1:
+        .long   0x12345678
+.LC2:
+        .long   0xbabebec0
+.LTN0_0:
+        l       %r1,.LC0-.LT0_0(%r13)
+        mvc     0(4,%r1),.LC1-.LT0_0(%r13)
+        l       %r1,.LC2-.LT1_0(%r13)
+        lr      %r13,%r0
+        br      %r1
+  */
+  /* What about big endian / little endian ?? */
+  *(unsigned short *) (function + 0) = 0x180D;
+  *(unsigned int *)   (function + 2) = 0xA7D50008;
+  *(unsigned int *)   (function + 6) = (unsigned int) variable;
+  *(unsigned int *)   (function +10) = (unsigned int) data;
+  *(unsigned int *)   (function +14) = (unsigned int) address;
+  *(unsigned int *)   (function +18) = 0x5810D000;
+  *(unsigned int *)   (function +22) = 0xD2031000;
+  *(unsigned short *) (function +26) = 0xD004;
+  *(unsigned int *)   (function +28) = 0x5810D008;
+  *(unsigned short *) (function +32) = 0x18D0;
+  *(unsigned short *) (function +34) = 0x07f1;
+#define is_tramp(function)  \
+  *(short *)          (function + 0) == 0x180D && \
+  *(int *)            (function + 2) == 0xA7D50008 && \
+  *(int *)            (function +18) == 0x5810D000 && \
+  *(int *)            (function +22) == 0xD2031000 && \
+  *(short *)          (function +26) == 0xD004 && \
+  *(int *)            (function +28) == 0x5810D008 && \
+  *(short *)          (function +32) == 0x18D0 && \
+  *(short *)          (function +34) == 0x07f1
+#define tramp_address(function)  \
+  *(unsigned int *) (function +14)
+#define tramp_variable(function)  \
+  *(unsigned int *) (function +6)
+#define tramp_data(function)  \
+  *(unsigned int *) (function +10)
+#endif
 
   /* 3. Set memory protection to "executable" */
 
@@ -1137,7 +1205,7 @@ __TR_function alloc_trampoline (address, variable, data)
    * cache. The freshly built trampoline is visible to the data cache, but not
    * maybe not to the instruction cache. This is hairy.
    */
-#if !(defined(__hppanew__) || defined(__rs6000aix__) || defined(__ia64__))
+#if !(defined(__hppanew__) || defined(__powerpcaix__) || defined(__powerpc64aix__) || defined(__ia64__))
   /* Only needed if we really set up machine instructions. */
 #ifdef __i386__
 #if defined(_WIN32)
@@ -1220,7 +1288,7 @@ __TR_function alloc_trampoline (address, variable, data)
 #ifdef __arm__
   /* This CPU does not have a separate instruction cache. (I think.) */
 #endif
-#ifdef __rs6000__
+#if defined(__powerpc__) && !defined(__powerpc64__)
   __TR_clear_cache(function);
 #endif
 #ifdef __m88k__
@@ -1235,12 +1303,7 @@ __TR_function alloc_trampoline (address, variable, data)
   return (__TR_function) (function + TRAMP_BIAS);
 }
 
-#if defined(__STDC__) || defined(__GNUC__) || defined(__cplusplus)
 void free_trampoline (__TR_function function)
-#else
-void free_trampoline (function)
-  __TR_function function;
-#endif
 {
 #if TRAMP_BIAS
   function = (__TR_function)((char*)function - TRAMP_BIAS);
@@ -1253,12 +1316,7 @@ void free_trampoline (function)
 #endif
 }
 
-#if defined(__STDC__) || defined(__GNUC__) || defined(__cplusplus)
 int is_trampoline (void* function)
-#else
-int is_trampoline (function)
-  void* function;
-#endif
 {
 #ifdef is_tramp
 #ifdef __hppanew__
@@ -1271,12 +1329,7 @@ int is_trampoline (function)
 #endif
 }
 
-#if defined(__STDC__) || defined(__GNUC__) || defined(__cplusplus)
 __TR_function trampoline_address (void* function)
-#else
-__TR_function trampoline_address (function)
-  void* function;
-#endif
 {
 #ifdef tramp_address
   return (__TR_function)(tramp_address(((char*)function - TRAMP_BIAS)));
@@ -1285,12 +1338,7 @@ __TR_function trampoline_address (function)
 #endif
 }
 
-#if defined(__STDC__) || defined(__GNUC__) || defined(__cplusplus)
 void* trampoline_variable (void* function)
-#else
-void* trampoline_variable (function)
-  void* function;
-#endif
 {
 #ifdef tramp_variable
   return (void*)(tramp_variable(((char*)function - TRAMP_BIAS)));
@@ -1299,12 +1347,7 @@ void* trampoline_variable (function)
 #endif
 }
 
-#if defined(__STDC__) || defined(__GNUC__) || defined(__cplusplus)
 void* trampoline_data (void* function)
-#else
-void* trampoline_data (function)
-  void* function;
-#endif
 {
 #ifdef tramp_data
   return (void*)(tramp_data(((char*)function - TRAMP_BIAS)));
