@@ -1,6 +1,6 @@
 ;;; Sources for CLISP DEFSTRUCT macro
 ;;; Bruno Haible 1988-2005
-;;; Sam Steingold 1998-2005
+;;; Sam Steingold 1998-2006
 ;;; German comments translated into English: Stefan Kain 2003-01-14
 
 (in-package "SYSTEM")
@@ -1195,13 +1195,13 @@
         (setf (clos::class-instance-size class) new-value)))))
 |#
 
-(defun structure-kconstructor (name)
+(defun structure-keyword-constructor (name)
   (let ((desc (get name 'DEFSTRUCT-DESCRIPTION)))
     (if desc
       (svref desc *defstruct-description-kconstructor-location*)
       (clos::class-kconstructor (find-class name)))))
 #|
- (defun (setf structure-kconstructor) (new-value name)
+ (defun (setf structure-keyword-constructor) (new-value name)
   (let ((desc (get name 'DEFSTRUCT-DESCRIPTION)))
     (if desc
       (setf (svref desc *defstruct-description-kconstructor-location*) new-value)
@@ -1253,7 +1253,7 @@
     (macrolet ((fmakunbound-if-present (symbol-form)
                  `(let ((symbol ,symbol-form))
                     (when symbol (fmakunbound symbol)))))
-      (fmakunbound-if-present (structure-kconstructor name))
+      (fmakunbound-if-present (structure-keyword-constructor name))
       (mapc #'fmakunbound (structure-boa-constructors name))
       (fmakunbound-if-present (structure-copier name))
       (fmakunbound-if-present (structure-predicate name))
