@@ -134,13 +134,15 @@ to print the corresponding values, or T for all of them.")
        (format stream (TEXT "a load-time evaluation promise.")))
       (EXT:SYMBOL-MACRO
        (format stream (TEXT "a symbol macro handler.")))
+      (EXT:GLOBAL-SYMBOL-MACRO
+       (format stream (TEXT "a global symbol macro handler.")))
       (SYS::MACRO
        (format stream (TEXT "a macro expander."))
-       (terpri stream)
-       (format stream (TEXT "For more information, evaluate ~{~S~^ or ~}.")
-               `((DISASSEMBLE (MACRO-FUNCTION
-                               ',(sys::closure-name
-                                  (sys::macro-expander obj)))))))
+       (let ((name (sys::closure-name (sys::macro-expander obj))))
+         (describe-arglist stream name)
+         (terpri stream)
+         (format stream (TEXT "For more information, evaluate ~{~S~^ or ~}.")
+                 `((DISASSEMBLE (MACRO-FUNCTION ',name))))))
       (EXT:FUNCTION-MACRO
        (format stream (TEXT "a function with alternative macro expander.")))
       (EXT:ENCODING
@@ -519,39 +521,27 @@ to print the corresponding values, or T for all of them.")
                                             (sys::%record-ref obj 3)
                                             (sys::%record-ref obj 4)))))
       (COMPILED-FUNCTION
-       (multiple-value-bind (name req opt rest-p keywords other-keys)
-           (sys::subr-info obj)
-         (if (and name req)
-           (progn
-             (format stream (TEXT "a built-in system function."))
-             (sys::describe-signature stream req opt rest-p
-                                      keywords keywords other-keys))
-           (progn
-             (format stream (TEXT "a compiled function."))
-             (multiple-value-bind (req opt rest-p key-p keywords other-keys-p)
-                 (sys::signature obj)
-               (sys::describe-signature stream req opt rest-p key-p keywords
-                                        other-keys-p)
-               (let* ((name (sys::closure-name obj))
-                      (funform (cond ((and (symbolp name) (macro-function name))
-                                      `(MACRO-FUNCTION ',name))
-                                     ((fboundp name) `(FUNCTION ,name)))))
-                 (when funform
-                   (terpri stream)
-                   (format stream
-                           (TEXT "For more information, evaluate ~{~S~^ or ~}.")
-                           `((DISASSEMBLE ,funform))))))))))
+       (let ((subrp (sys::subr-info obj)))
+         (format stream (if subrp
+                            (TEXT "a built-in system function.")
+                            (TEXT "a compiled function.")))
+         (describe-arglist stream obj)
+         (describe-documentation stream obj)
+         (let* ((name (sys::function-name obj))
+                (funform (cond ((and (symbolp name) (macro-function name))
+                                `(MACRO-FUNCTION ',name))
+                               ((fboundp name) `(FUNCTION ,name)))))
+           (when funform
+             (terpri stream)
+             (format stream
+                     (TEXT "For more information, evaluate ~{~S~^ or ~}.")
+                     `((DISASSEMBLE ,funform)))))))
       (FUNCTION
        ;; we do not use ETYPECASE here to ensure that if we do get here,
        ;; we are dealing with an Iclosure object
        (format stream (TEXT "an interpreted function."))
-       (let ((doc (sys::%record-ref obj 2)))
-         (terpri stream)
-         (format stream (TEXT "Argument list: ~:S")
-                 (car (sys::%record-ref obj 1)))
-         (when doc
-           (terpri stream)
-           (format stream (TEXT "Documentation: ~A") doc)))))))
+       (describe-arglist stream obj)
+       (describe-documentation stream obj)))))
 
 (defun describe1 (obj stream)
   (let ((objstring (sys::write-to-short-string
@@ -559,10 +549,12 @@ to print the corresponding values, or T for all of them.")
     (fresh-line stream)
     (terpri stream) ; blank line
     (if (memq obj *describe-done*)
-      (format stream (TEXT "~A [see above]") objstring)
+      (with-stream-s-expression (stream)
+        (format stream (TEXT "~A [see above]") objstring))
       (let ((doc (and (symbolp obj) (get obj 'sys::doc))))
         (push obj *describe-done*)
-        (format stream (TEXT "~A is ") objstring)
+        (with-stream-s-expression (stream)
+          (format stream (TEXT "~A is ") objstring))
         (describe-object obj stream)
         (fresh-line stream)
         (when doc
@@ -604,20 +596,32 @@ to print the corresponding values, or T for all of them.")
 ;; auxiliary functions for DESCRIBE of FUNCTION
 
 (defun arglist (func)
-  (setq func (coerce func 'function))
-  (if (typep func 'generic-function)
-    ; Generic functions store the lambda-list. It has meaningful variable names.
-    (clos:generic-function-lambda-list func)
-    ; Normal functions store only the signature, no variable names.
-    (sig-to-list (get-signature func))))
+  (cond ((and (symbolp func) (fboundp func) 
+              (sys::macrop (symbol-function func)))
+         (sys::macro-lambda-list (symbol-function func)))
+        ((typep (setq func (coerce func 'function)) 'generic-function)
+         ;; Generic functions store the original lambda-list.
+         (clos:generic-function-lambda-list func))
+        ((or (sys::subr-info func) ; built-in
+             #+FFI (eq (type-of func) 'FFI::FOREIGN-FUNCTION))
+         (sig-to-list (get-signature func)))
+        ((sys::%compiled-function-p func) ; compiled closure
+         (or (sys::closure-lambda-list func)
+             (sig-to-list (get-signature func))))
+        ((sys::closurep func) ; interpreted closure?
+         (car (sys::%record-ref func 1)))))
 
-(defun describe-signature (stream req-anz opt-anz rest-p keyword-p keywords
-                           allow-other-keys)
+(defun describe-arglist (stream function)
   (terpri stream)
-  (format stream (TEXT "Argument list: ~A.")
-          (format nil "(~{~A~^ ~})"
-                  (signature-to-list req-anz opt-anz rest-p keyword-p keywords
-                                     allow-other-keys))))
+  (format stream (TEXT "Argument list: ~:S")
+          (handler-case (arglist function)
+            (error (c) (princ-to-string c)))))
+
+(defun describe-documentation (stream function)
+  (let ((doc (clos::function-documentation function)))
+    (when doc
+      (terpri stream)
+      (format stream (TEXT "Documentation: ~A") doc))))
 
 ;;-----------------------------------------------------------------------------
 ;; auxiliary functions for CLISP metadata
