@@ -1,6 +1,6 @@
 ;;; Foreign function interface for CLISP
 ;;; Bruno Haible 19.2.1995
-;;; Sam Steingold 1998-2005
+;;; Sam Steingold 1998-2007
 
 #+UNICODE
 (progn
@@ -17,7 +17,7 @@
           c-lines *output-c-functions* *output-c-variables* *foreign-guard*
           nil boolean character char uchar short ushort int uint long ulong
           uint8 sint8 uint16 sint16 uint32 sint32 uint64 sint64
-          single-float double-float
+          single-float double-float default-foreign-library
           c-pointer c-string c-struct c-union c-array c-array-max
           c-function c-ptr c-ptr-null c-array-ptr
           def-c-enum def-c-struct element deref slot cast typeof
@@ -353,6 +353,13 @@
   `(eval-when (load compile eval)
      (without-package-lock ("FFI") (setq *foreign-language* ',lang))))
 
+;; the default foreign library for this compilation unit
+(defvar *foreign-library* nil) ; ABI
+
+(defmacro default-foreign-library (library)
+  `(eval-when (load compile eval)
+     (without-package-lock ("FFI") (setq *foreign-library* ,library))))
+
 ;; get the even (start=0) or odd (start=1) elements of the simple vector
 (defun split-c-fun-arglist (args start)
   (do ((ii start (+ ii 2)) (res '()))
@@ -423,10 +430,14 @@
     (error (TEXT "The name ~S is not a valid C identifier")
            name)))
 
-(defmacro DEF-C-TYPE (&whole whole-form name typespec)
+(defmacro DEF-C-TYPE (&whole whole-form name &optional typespec)
   (setq name (check-symbol name (first whole-form)))
   `(EVAL-WHEN (LOAD COMPILE EVAL)
-     (PARSE-C-TYPE ',typespec ',name)
+     ,(if typespec
+          `(PARSE-C-TYPE ',typespec ',name)
+          `(SETF (GETHASH ',name FFI::*C-TYPE-TABLE*)
+                 ;; prefer lower case
+                 (CS-CL:symbol-name ',name)))
      ',name))
 
 ;; Convert back a C type from internal (vector) to external (list)
@@ -758,11 +769,23 @@
             ~%void module__~A__init_function_2 (module_t* module);~%~
             ~%void module__~A__fini_function (module_t* module);~%~
             ~2%void module__~A__init_function_1 (module_t* module)~%~
-            {~{~%~A~}~%}~2%~
-            void module__~A__init_function_2 (module_t* module)~%~
             {~{~%~A~}~%"
-            *c-name* *c-name* *c-name*
-            *c-name* *init-once* *c-name* *init-always*)
+            *c-name* *c-name* *c-name* *c-name* *init-once*)
+    (let ((done (make-hash-table :test 'equal)))
+      (maphash (lambda (type spec)
+                 (declare (ignore type))
+                 (when (and (stringp spec) (not (gethash spec done)))
+                   (setf (gethash spec done) spec)
+                   (when *foreign-guard*
+                     (format *coutput-stream* "# if HAVE_~A~%"
+                             (string-upcase spec)))
+                   (format *coutput-stream* "  register_foreign_inttype(~S,sizeof(~A),(~A)-1<=(~A)0);~%" spec spec spec spec)
+                   (when *foreign-guard* (format *coutput-stream* "# endif~%"))))
+               *c-type-table*))
+    (format *coutput-stream*
+            "}~2%void module__~A__init_function_2 (module_t* module)~%~
+            {~{~%~A~}~%"
+             *c-name* *init-always*)
     (dolist (variable *variable-list*)
       (let ((c-name (first variable)))
         (when *foreign-guard*
@@ -877,6 +900,10 @@
                 'def-c-const name c-name)
           (sys::%unbound)))))
 
+(defun get-assoc (key alist default)
+  (let ((pair (assoc key alist)))
+    (if pair (second pair) default)))
+
 (defmacro DEF-C-VAR (&whole whole-form
                      name &rest options)
   (setq name (check-symbol name (first whole-form)))
@@ -899,19 +926,15 @@
                          (:NONE 0)
                          (:MALLOC-FREE fv-flag-malloc-free))
                        0))))
-         (library  (second (assoc :library alist)))
+         (library (get-assoc :library alist '*foreign-library*))
          #|
          (getter-function-name (sys::symbol-suffix name "%GETTER%"))
          (setter-function-name (sys::symbol-suffix name "%SETTER%"))
          |#
          (def (gensym "DEF-C-VAR-")))
     `(LET ((,def (LOAD-TIME-VALUE
-                  ,(if library
-                       `(FFI::FOREIGN-LIBRARY-VARIABLE
-                         ',c-name (FFI::FOREIGN-LIBRARY ,library)
-                         nil (PARSE-C-TYPE ',type))
-                       `(LOOKUP-FOREIGN-VARIABLE
-                         ',c-name (PARSE-C-TYPE ',type))))))
+                  (FIND-FOREIGN-VARIABLE
+                   ',c-name (PARSE-C-TYPE ',type) ,library NIL))))
        #|
        (LET ((FVAR (LOOKUP-FOREIGN-VARIABLE ',c-name (PARSE-C-TYPE ',type))))
          (DEFUN ,getter-function-name () (FOREIGN-VALUE FVAR))
@@ -922,8 +945,8 @@
        (DEFSETF ,getter-function-name ,setter-function-name)
        (DEFINE-SYMBOL-MACRO ,name (,getter-function-name))
        |#
-       ,(unless library
-          `(EVAL-WHEN (COMPILE) (NOTE-C-VAR ',c-name ',type ',flags)))
+       (EVAL-WHEN (COMPILE)
+         (UNLESS ,LIBRARY (NOTE-C-VAR ',c-name ',type ',flags)))
        (when ,def
          (SYSTEM::%PUT ',name 'FOREIGN-VARIABLE ,def)
          ,@(when doc `((SETF (DOCUMENTATION ',name 'VARIABLE) ',(second doc))))
@@ -1026,26 +1049,25 @@
                                    :built-in :library :documentation)
                          whole-form))
          (def (gensym "DEF-CALL-OUT-"))
-         (doc (assoc ':documentation alist))
-         (library (second (assoc :library alist)))
+         (properties (and (>= 1 (compiler::declared-optimize
+                                 'space (and (boundp 'system::*denv*)
+                                             system::*denv*)))
+                          (assoc ':documentation alist)))
+         (library (get-assoc :library alist '*foreign-library*))
          (c-name (foreign-name name (assoc :name alist)))
          (built-in (second (assoc :built-in alist)))
          ;; Maximize sharing in .fas file, reuse options
          ;; parse-c-function ignores unknown options, e.g. :name
          (ctype `(PARSE-C-FUNCTION ',options ',whole-form)))
-    `(LET ((,def ,(if library
-                      `(FFI::FOREIGN-LIBRARY-FUNCTION
-                        ',c-name (FFI::FOREIGN-LIBRARY ,library)
-                        NIL ,ctype)
-                      `(LOOKUP-FOREIGN-FUNCTION ',c-name ,ctype))))
+    `(LET ((,def (FIND-FOREIGN-FUNCTION
+                  ',c-name ,ctype ',properties ,library NIL)))
        (EXT:COMPILER-LET ((,def ,ctype))
-         ,(unless library
-            `(EVAL-WHEN (COMPILE) (NOTE-C-FUN ',c-name ,def ',built-in)))
+         (EVAL-WHEN (COMPILE)
+           (UNLESS ,LIBRARY (NOTE-C-FUN ',c-name ,def ',built-in)))
          (COMPILER::EVAL-WHEN-COMPILE
            (COMPILER::C-DEFUN ',name (C-TYPE-TO-SIGNATURE ,ctype))))
        (WHEN ,def                       ; found library function
          (SYSTEM::REMOVE-OLD-DEFINITIONS ',name)
-         ,@(when doc `((SETF (DOCUMENTATION ',name 'FUNCTION) ',(second doc))))
          (SYSTEM::%PUTD ',name ,def))
        ',name)))
 
@@ -1061,6 +1083,10 @@
   (let* ((alist (parse-options options
                                '(:name :offset :arguments :return-type)
                                whole-form))
+         (properties (and (>= 1 (compiler::declared-optimize
+                                 'space (and (boundp 'system::*denv*)
+                                             system::*denv*)))
+                          (assoc ':documentation alist)))
          (c-name (foreign-name name (assoc ':name alist)))
          (offset (second (assoc ':offset alist)))
          (ctype `(PARSE-C-FUNCTION ',options ',whole-form)))
@@ -1068,10 +1094,8 @@
        (SYSTEM::REMOVE-OLD-DEFINITIONS ',name)
        (COMPILER::EVAL-WHEN-COMPILE
          (COMPILER::C-DEFUN ',name (C-TYPE-TO-SIGNATURE ,ctype)))
-       (SYSTEM::%PUTD ',name
-         (FOREIGN-LIBRARY-FUNCTION ',c-name
-           (FOREIGN-LIBRARY ',library)
-           ',offset ,ctype))
+       (SYSTEM::%PUTD ',name (FIND-FOREIGN-FUNCTION
+                              ',c-name ,ctype ',properties ',library ',offset))
        ',name)))
 
 (defun count-inarguments (arg-vector)
@@ -1108,9 +1132,7 @@
        ',name)))
 
 (defun convert-to-foreign-C (flags)
-  (if (flag-set-p flags ff-flag-malloc-free)
-      "convert_to_foreign_mallocing"
-      "convert_to_foreign_nomalloc"))
+  (if (flag-set-p flags ff-flag-malloc-free) "mallocing" "nomalloc"))
 
 (defun note-c-call-in (name c-name alist whole) ; ABI
   (when (compiler::prepare-coutput-file)
@@ -1159,28 +1181,28 @@
         (format *coutput-stream* "  funcall(~A,~D);~%"
                 (object-to-c-value (pass-object name)) inargcount)
         (unless (eq rettype 'NIL)
-          (format *coutput-stream* " {~%  var ~A;~%~:
-  ~A(~A,value1,&retval);~%"
+          (format *coutput-stream* " {~%  ~A;~%~:
+  convert_to_foreign(~A,value1,&retval,&~A);~%"
                   (to-c-typedecl rettype "retval")
-                  (convert-to-foreign-C flags)
-                  (object-to-c-value (pass-object rettype))))
+                  (object-to-c-value (pass-object rettype))
+                  (convert-to-foreign-C flags)))
         (let ((outargcount (if (eq rettype 'NIL) 0 1)))
           (mapc #'(lambda (argtype argflag argname)
                     (when (flag-set-p argflag flag-output)
                       (unless (eq (ctype-type argtype) 'C-PTR)
                         (error (TEXT "~S: :OUT argument is not a pointer: ~S")
                                'DEF-CALL-IN argtype))
-                      (format *coutput-stream* "  ~A~A(~A,~A,~A);~%"
+                      (format *coutput-stream*
+                              "  ~Aconvert_to_foreign(~A,~A,~A,&~A);~%"
                               (if (eql outargcount 0) ""
                                 (format nil "if (mv_count >= ~D) "
                                         (+ outargcount 1)))
-                              (convert-to-foreign-C argflag)
                               (object-to-c-value
                                 (pass-object (svref argtype 1)))
                               (if (eql outargcount 0)
                                 "value1"
                                 (format nil "mv_space[~D]" outargcount))
-                              argname)
+                              argname (convert-to-foreign-C argflag))
                       (incf outargcount)))
                 argtypes argflags argnames))
         (format *coutput-stream* "  end_callback();~%")
