@@ -1,7 +1,7 @@
 ;; Tracer
 ;; Bruno Haible 13.2.1990, 15.3.1991, 4.4.1991
 ;; German comments translated into English: Stefan Kain 2001-12-26
-;; Sam Steingold 2001-2002
+;; Sam Steingold 2001-2007
 
 ;; (TRACE) returns a list of traced functions
 ;; (TRACE fun ...) additionally traces the functions fun, ... .
@@ -11,6 +11,7 @@
 ;;  or a List made of a Symbol and a few Keyword-Arguments (pair-wise!)
 ;;   (symbol
 ;;    [:suppress-if form]   ; no Trace-Output, as long as form is true
+;;    [:max-depth form]     ; no trace output, as long as (> *trace-level* form)
 ;;    [:step-if form]       ; Trace moves into the Stepper, if form is true
 ;;    [:pre form]           ; executes form before function call
 ;;    [:post form]          ; executes form after  function call
@@ -47,7 +48,7 @@
   "Use indentation in addition to numbering to indicate the trace level.")
 
 (proclaim '(special *trace-function* *trace-args* *trace-form* *trace-values*))
-(defvar *traced-functions* nil) ; list of currently traced function-names
+(defvar *traced-functions* nil) ; list of currently traced function-names ; ABI
 ;; So long as a function-name funname [resp. more exactly: the Symbol
 ;; symbol = (get-funname-symbol funname)] are traced, the Property
 ;; sys::traced-definition contains the old content of the function-cell, the
@@ -60,6 +61,7 @@
 ;;       (eq (symbol-function symbol) (get symbol 'sys::tracing-definition)))
 ;; ===>   (member funname *traced-functions* :test #'equal)
 ;; <==>   (get symbol 'sys::traced-definition)
+
 (defvar *trace-level* 0) ; nesting depth for Trace-Output
 
 (labels ((subclosure-pos (closure name)
@@ -81,8 +83,9 @@
              (unless (closurep closure)
                (error-of-type 'type-error
                  :datum closure :expected-type 'closure
-                 (TEXT "~S: ~S must name a closure") 'local name))
-             (if (compiled-function-p closure) closure
+                 (TEXT "~S: ~S does not name a closure") 'local name))
+             (if (sys::%compiled-function-p closure)
+                 closure
                  (fdefinition (compile name closure)))))
          (local-helper (spec)
            (do* ((spe (cdr spec) (cdr spe))
@@ -92,17 +95,18 @@
                       (subclosure-pos clo (car spe))))
                 ((endp (cdr spe)) (values clo pos)))))
 
-  (defun %local-get (spec)
+  (defun %local-get (spec) ; ABI
     (multiple-value-bind (clo pos) (local-helper spec)
       (sys::%record-ref clo pos)))
-  (defun %local-set (new-def spec)
+  (defun %local-set (new-def spec) ; ABI
     (unless (closurep new-def)
       (error-of-type 'type-error
         :datum new-def :expected-type 'closure
-        (TEXT "~S: ~S must be a closure") `(setf (local ,@spec) ,new-def)))
+        (TEXT "~S: ~S is not a closure") `(setf (local ,@spec)) new-def))
     (multiple-value-bind (clo pos) (local-helper spec)
       (sys::%record-store clo pos
-         (if (compiled-function-p new-def) new-def
+         (if (sys::%compiled-function-p new-def)
+             new-def
              (fdefinition (compile (closure-name (sys::%record-ref clo pos))
                                    new-def)))))))
 
@@ -116,16 +120,19 @@ CLOSURE must be compiled."
   "Modify the local definition (LABELS or FLET).
 This will not work with closures that use lexical variables!"
   (let ((store (gensym "LOCAL-")))
-    (values nil nil `(,store) `(%local-set ,store ',spec)
-            `(%local-get ,spec))))
+    (values nil nil `(,store) `(%LOCAL-SET ,store ',spec)
+            `(%LOCAL-GET ,spec))))
 
 ;; check whether the object might name a local (LABELS or FLET) function
 (defun local-function-name-p (obj)
   (and (consp obj) (eq 'local (car obj))))
 
+;; Structure containing all trace options for a given function.
+;; make-tracer is ABI
 (defstruct (tracer (:type vector))
   name symb cur-def local-p
-  suppress-if step-if pre post pre-break-if post-break-if
+  suppress-if max-depth step-if pre post pre-break-if post-break-if
+  bindings
   pre-print post-print print)
 
 ;; install the new function definition
@@ -136,25 +143,14 @@ This will not work with closures that use lexical variables!"
 
 (defmacro trace (&rest funs)
   (if (null funs)
-    '*traced-functions*
-    (cons 'append
-      (mapcar (lambda (fun)
-                (cond ((or (function-name-p fun) (local-function-name-p fun))
-                       (let ((trr (make-tracer :name fun)))
-                         (check-traceable fun trr 'trace)
-                         `(trace1 ,trr)))
-                      ((let ((trr (apply #'make-tracer :name fun)))
-                         (check-traceable (first fun) trr 'trace)
-                         `(trace1 ,trr)))))
-              funs))))
-
-(defun check-function-name (caller funname)
-  (do () ((function-name-p funname) funname)
-    (setq funname (check-value nil (coerce-to-condition
-                                    (TEXT "~s: ~s is not a function name")
-                                    (list caller funname)
-                                    'check-function-name
-                                    'simple-source-program-error)))))
+    '*TRACED-FUNCTIONS*
+    `(APPEND
+      ,@(mapcar #'(lambda (fun)
+                    `(TRACE1
+                       ,(if (or (function-name-p fun) (local-function-name-p fun))
+                          `(MAKE-TRACER :NAME ',fun)
+                          `(APPLY #'MAKE-TRACER :NAME ',fun))))
+                funs))))
 
 ;; check whether the FUNNAME can be traced,
 ;; fill SYMB, CUR-DEF and LOCAL-P slots of TRR and return TRR
@@ -177,12 +173,13 @@ This will not work with closures that use lexical variables!"
            (when (get (tracer-symb trr) 'sys::untraced-name)
              (setf (tracer-symb trr)
                    (get (tracer-symb trr) 'sys::untraced-name))))
-          (t (setq funname (check-function-name caller funname))
+          (t (setq funname (check-function-name funname caller))
              (go restart))))
-  (check-redefinition funname caller "function")
+  (check-redefinition funname caller (TEXT "function"))
   trr)
 
-(defun trace1 (trr)
+(defun trace1 (trr) ; ABI
+  (check-traceable (tracer-name trr) trr 'trace)
   (let ((macro-flag (macrop (tracer-cur-def trr)))
         (sig (when (tracer-local-p trr)
                (sig-to-list (get-signature (tracer-cur-def trr))))))
@@ -191,16 +188,20 @@ This will not work with closures that use lexical variables!"
       (setf (get (tracer-symb trr) 'sys::traced-definition)
             (tracer-cur-def trr))
       (pushnew (tracer-name trr) *traced-functions* :test #'equal))
-    (format t (TEXT "~&;; Tracing ~:[function~;macro~] ~S.")
+    (fresh-line)
+    (format t (TEXT ";; Tracing ~:[function~;macro~] ~S.")
             macro-flag (tracer-name trr))
     (setf (get (tracer-symb trr) 'sys::tracing-definition)
           ;; new function, that replaces the original one:
           (let ((newname (concat-pnames "TRACED-" (tracer-symb trr)))
                 (body
                  `((declare (inline car cdr cons apply values-list))
-                   (let ((*trace-level* (1+ *trace-level*)))
+                   (let ((*trace-level* (1+ *trace-level*))
+                         ,@(tracer-bindings trr))
                      (block nil
-                       (unless ,(tracer-suppress-if trr) (trace-pre-output))
+                       (unless (or ,(tracer-suppress-if trr)
+                                   ,(if (tracer-max-depth trr) `(> *trace-level* ,(tracer-max-depth trr)) 'nil))
+                         (trace-pre-output))
                        ,@(when (tracer-pre-print trr)
                            `((trace-print (multiple-value-list
                                            ,(tracer-pre-print trr)))))
@@ -226,11 +227,12 @@ This will not work with closures that use lexical variables!"
                              `((trace-print (multiple-value-list
                                              ,(tracer-print trr)))))
                          ,@(when (tracer-post-print trr)
-                                `((trace-print (multiple-value-list
+                             `((trace-print (multiple-value-list
                                              ,(tracer-post-print trr)))))
-                         (unless ,(tracer-suppress-if trr)
-                               (trace-post-output))
-                             (values-list *trace-values*)))))))
+                         (unless (or ,(tracer-suppress-if trr)
+                                     ,(if (tracer-max-depth trr) `(> *trace-level* ,(tracer-max-depth trr)) 'nil))
+                           (trace-post-output))
+                         (values-list *trace-values*)))))))
             (setf (get newname 'sys::untraced-name) (tracer-symb trr))
             (macrolet ((f (def) `(fdefinition (compile newname ,def))))
               (cond (macro-flag
@@ -241,7 +243,8 @@ This will not work with closures that use lexical variables!"
                                     (macro-expander
                                      (get-traced-definition
                                       ',(tracer-symb trr)))))
-                            ,@body))))
+                            ,@body))
+                      (macro-lambda-list (tracer-cur-def trr))))
                     ((tracer-local-p trr)
                      (f `(lambda ,sig
                            (let* ((*trace-args* (list ,@sig))
@@ -280,38 +283,38 @@ This will not work with closures that use lexical variables!"
                     (cons 'quote (cons arg nil)))
                 args)))
 ;; Output before call, uses *trace-level* and *trace-form*
-(defun trace-pre-output ()
-  (terpri *trace-output*)
+(defun trace-output () ; common for pre & post
+  (fresh-line *trace-output*)
   (when *trace-indent*
     (write-spaces *trace-level* *trace-output*))
   (write *trace-level* :stream *trace-output* :base 10 :radix t)
-  (write-string " Trace: " *trace-output*)
-  (prin1 *trace-form* *trace-output*))
+  (write-string " Trace: " *trace-output*))
+(defun trace-pre-output ()
+  (trace-output)
+  (prin1 *trace-form* *trace-output*)
+  (elastic-newline *trace-output*))
 ;; Output after call, uses *trace-level*, *trace-form* and *trace-values*
 (defun trace-post-output ()
   (declare (inline car cdr consp atom))
-  (terpri *trace-output*)
-  (when *trace-indent*
-    (write-spaces *trace-level* *trace-output*))
-  (write *trace-level* :stream *trace-output* :base 10 :radix t)
-  (write-string " Trace: " *trace-output*)
+  (trace-output)
   (write (car *trace-form*) :stream *trace-output*)
   (write-string " ==> " *trace-output*)
   (trace-print *trace-values* nil))
 ;; Output of a list of values:
 (defun trace-print (vals &optional (nl-flag t))
-  (when nl-flag (terpri *trace-output*))
+  (when nl-flag (fresh-line *trace-output*))
   (when (consp vals)
     (loop
       (prin1 (pop vals) *trace-output*)
       (when (atom vals) (return))
-      (write-string ", " *trace-output*))))
+      (write-string ", " *trace-output*)))
+  (elastic-newline *trace-output*))
 
 (defmacro untrace (&rest funs)
-  `(mapcan #'untrace1
-    ,(if (null funs) `(copy-list *traced-functions*) `',funs)))
+  `(MAPCAN #'UNTRACE1
+     ,(if (null funs) `(COPY-LIST *TRACED-FUNCTIONS*) `',funs)))
 
-(defun untrace1 (funname)
+(defun untrace1 (funname) ; ABI
   (let* ((trr (check-traceable funname (make-tracer :name funname) 'untrace))
          (symbol (tracer-symb trr))
          (old-definition (get symbol 'sys::traced-definition)))
